# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from . import internal
from . import er


def add_deploy_https_key_cmd(s):
    p = s.add_parser(
        'deployHttpsKey', help='申请并部署 https 证书。')
    internal.add_arg_host(p)
    p.add_argument("-f", '--force',
                   help="重新申请证书。", default=False, action='store_true')
    p.set_defaults(func=deploy_https_key_cmd)


def deploy_https_key_cmd(p):
    host = internal.get_arg_host_from_conf(p)
    b = 'exec -T acme.sh redeploy-ssl'
    if p.force:
        b += ' --force'
    # redeploy-ssl 是作为文件映射进去的，需要重启才会更新。
    internal.docker_compose_cmd('restart acme.sh', host)
    print("开始申请证书，如果出现文件复制失败，请确认 nginx 容器是否正常运行。")
    return internal.docker_compose_cmd(b, host)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

import os
from . import internal
from . import er


def add_backup_cmd(s):
    p = s.add_parser(
        'backup', help='基于 rsync --del --link-dest 的增量备份。更改参数前先测试，以免写错删掉已有的备份文件。')
    internal.add_arg_host(p)
    internal.add_arg_force(p)
    p.set_defaults(func=backup_cmd)
    p.add_argument('obj', type=str, default='volumes', choices=[
        'docker',  'release', 'servers', 'var', 'volumes', 'ops', 'all'], nargs='?',
        help='''同步的文件夹。docker: docker-compose，ops 是除 var 和 volumes 的所有。''')
    p.add_argument('-t', '--target',
                   help="目标路径", default='backup/')
    p.add_argument('-d', '--time-format', default='%Y-%m-%d_%H:%M:%S',
                   help='目标路径下创建文件夹名的时间模板，与 python time.strftime format 参数相同。如 %%Y-%%m-%%d_%%H:%%M:%%S')
    p.add_argument('-l', '--link-dest',
                   help='未更改时链接到指定文件夹。默认是备份路径中符合 format 排序后最大的文件夹。')
    p.add_argument('-k', '--keep',
                   type=int, help="保留的备份个数，只有 format 开启时启用。本次备份也算一个。-1 保留所有。", default='-1')
    p.add_argument("-c", '--cod',
                   help="在目标路径创建项目名命名的文件夹", default=False, action='store_true')


def backup_cmd(p):
    host = internal.get_arg_host_from_conf(p)
    if not os.path.isdir(p.target):
        os.mkdir(p.target)
    return internal.backup(host,  p.obj, p.target, p.time_format, p.link_dest, p.keep, p.cod, p.force)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from . import internal


def add_nginx_reload_cmd(s):
    p = s.add_parser(
        'nginxReload', help='重载 nginx 配置，不会重载证书。')
    internal.add_arg_host(p)
    p.set_defaults(func=nginx_reload_cmd)


def nginx_reload_cmd(p):
    host = internal.get_arg_host_from_conf(p)
    return internal.docker_compose_cmd('exec -T nginx nginx -s reload', host)



#!/bin/env python3

# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

import argparse
import sys

desc = '''drops 是基于 ssh 和 docker-compose 的运维模板，附带的 drops 命令可以方便的管理项目，部署服务。'''

# 是否打印异常跟踪


def main():
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('--debug',
                        help="启动 debug 模式。", default=False, action='store_true')
    # metavar 设置空字符串，为了不以  {cmd1, cmd2, ..} 的形式显示可用子命令，怪怪的。
    subparsers = parser.add_subparsers(metavar="")

    internal.initCmd(subparsers)

    arg = parser.parse_args()
    # 调用相关命令，没有命令时打印 help
    if 'func' in arg:
        if not arg.debug:
            try:
                arg.func(arg)
            except Exception as e:
                print('Fatal:', type(e).__name__, ':', e)
                sys.exit(1)
        else:
            arg.func(arg)
    else:
        parser.print_help()


if __name__ == "__main__":
    import internal  # 为了单个文件调试
    main()
else:
    from . import internal  # 同时又能作为包被引入



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

import os
from . import internal


def add_sync_cmd(s):
    p = s.add_parser(
        'sync', help='同步当前项目到远程路径')
    internal.add_arg_host(p)
    internal.add_arg_force(p)
    p.set_defaults(func=sync_cmd)
    p.add_argument('obj', type=str, default='ops', choices=[
        'docker',  'release', 'servers', 'var', 'volumes', 'ops'], nargs='?',
        help='''同步的对象。docker: docker-compose，ops 是除 var 和 volumes 的所有。默认：ops。
var, volumes 建议只用来同步初始数据。
同步 var, volumes 会检查远程目录是否为空文件夹，并做相应提示。''')


def sync_cmd(p):
    host = internal.get_arg_host_from_conf(p)
    return internal.sync(host, p.force, p.obj)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import os
import shutil
from . import internal

import drops
from . import er


def add_clean_up_cmd(s):
    p = s.add_parser(
        'clean', help='删除当前目录下 drops 所有相关文件。')
    internal.add_arg_force(p)
    p.set_defaults(func=new_clean_up)


def new_clean_up(p):
    if not p.force and not internal.user_confirm('是否清理掉当前目录的 drops 相关文件？'):
        raise er.UserCancel
    objPath = os.path.join(drops.__path__[0], 'docker_ops')
    pwd = os.getcwd()
    if not os.path.isfile(os.path.join(pwd, 'drops.yaml')):
        raise er.ThisIsNotDropsProject()
    for i in os.listdir(objPath):
        p = os.path.join(objPath, i)
        t = os.path.join(pwd, i)
        if os.path.isdir(p):
            shutil.rmtree(t)
        else:
            os.remove(t)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


from . import er
from . import config


def host_cmd(a):
    if a.cmd == 'add' or a.cmd == 'change':
        if a.hostAlias == None:
            raise er.ArgsError('至少需要传入 hostAlias 参数')
        if a.host == None:
            raise er.ArgsError('至少需要传入 host 参数')
        c = {'host': a.host, 'port': a.port,
             'username': a.username, 'coding': a.coding, 'hostAlias': a.hostAlias}
        if a.password is not None:
            c['password'] = a.password
        elif a.key is not None:
            c['key'] = a.key
        else:
            raise er.ArgsError('密码或者 key 至少要传入一个。')
        if a.cmd == 'change':
            config.Conf().change_host(**c)
        else:
            config.Conf().add_host(**c)

    elif a.cmd == 'remove':
        if a.hostAlias == None:
            raise er.ArgsError('至少需要传入 hostAlias 参数')
        config.Conf().remove_host(a.hostAlias)
    elif a.cmd == 'ls':
        config.Conf().ls()
    else:
        config.Conf().ls()


def add_host_cmd(s):
    p = s.add_parser(
        'host', help='管理 drops 部署主机。因为密码是明文存储的，强烈建议用 key 做验证。')
    p.add_argument('cmd', type=str, choices=[
                   'ls', 'add', 'remove', 'change'], nargs='?', help="default ls")
    p.add_argument("hostAlias", metavar="default",
                   type=str, help="host 类型，比方说 default、test、dev、online。default 是关键字，所有命令不指定 --hostAlias 的话默认对 default 进行操作。")
    p.add_argument("host", metavar="ssh.example.com",
                   type=str, help="host.", default=None, nargs='?')
    p.add_argument("port", metavar="22",
                   type=int, help="ssh port.", default=22, nargs='?')
    p.add_argument("username", metavar="root",
                   type=str, help="ssh username.", default='root', nargs='?')
    p.add_argument('coding', metavar="coding",
                   type=str, help="shell coding, default is utf-8.", default='utf-8', nargs='?')
    p.add_argument("-p", '--password',
                   type=str, help="ssh password. Note that the password is stored in plain text, and key verification is strongly recommended.", default=None)
    p.add_argument("-k", '--key', metavar="~/.ssh/id_ed25519",
                   type=str, help="ssh key path", default=None)
    p.set_defaults(func=host_cmd)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from . import internal
from . import er


def add_undeploy_cmd(s):
    p = s.add_parser(
        'undeploy', help='清理掉服务器上的容器和项目')
    internal.add_arg_host(p)
    internal.add_arg_force(p)
    p.set_defaults(func=undeploy_cmd)


def undeploy_cmd(p):
    if not p.force and not internal.user_confirm('即将进行反部署，这会清理掉服务器上的容器及 '+internal.container_path()+' 目录，但不会完全删除映射文件。是否继续？'):
        raise er.UserCancel
    host = internal.get_arg_host_from_conf(p)
    print('---------- kill ----------')
    internal.docker_compose_cmd('kill', host)
    print('--------- rm -f ---------')
    internal.docker_compose_cmd('rm -f', host)
    print('-------- rm -rf %s --------' % internal.container_path())
    internal.exec('rm -rf %s' % internal.container_path(), host)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from paramiko.client import SSHClient
import paramiko
import os

from . import er


class Client():
    def __init__(self, host, port=22, username='root', password=None, key=None, coding='utf-8'):
        self._client = SSHClient()
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if 'known_hosts' in os.listdir('.'):
            self._client.load_system_host_keys('known_hosts')
        if password != None:
            self._client.connect(
                host, port=port, username=username, password=password)
        elif key != None:
            self._client.connect(
                host, port=port, username=username, key_filename=key)
        else:
            raise er.PwdAndKeyCannotBeEmpty

        self.coding = coding

    def exec(self, b, show=True):
        # 返回标准输出，标准错误，状态码
        # 没有 exit 0 的话会导致命令执行随机返回 -1，貌似是 paramiko 的锅，没找到具体原因。
        # 为了做出类似 ssh 交互式的效果，混合了标准输出和标准错误。并在这一层做输出，屏蔽复杂度。
        # 2>&1 重定向不管用，只能 exec 重定向
        stdin, stdout, stderr = self._client.exec_command(
            'exec 2>/dev/stdout && ' + b + ' && exit 0')
        so = ''
        # 命令执行过程可能很长，所以在这里按行处理输出。
        if show:
            while True:
                s = stdout.readline()
                if s == '':
                    break
                print(s, end='')
                so += s
        else:
            while True:
                s = stdout.read()
                if s == b'':
                    break
                so += s
        return so, stdout.channel.recv_exit_status()

    def exed(self, b):
        # 返回标准输出，标准错误，状态码
        # 没有 exit 0 的话会导致命令执行随机返回 -1，貌似是 paramiko 的锅，没找到具体原因。
        stdin, stdout, stderr = self._client.exec_command(
            b + ' exit 0')
        o, e, s = stdout.read().decode(self.coding), stderr.read().decode(
            self.coding), stdout.channel.recv_exit_status()
        return o, e, s



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import os
import time
import shutil

from . import er
from . import ssh
from . import config

port = 22  # 默认端口


def work_path():
    # 当前 drops 项目绝对路径，drops.yaml 所在目录
    return os.getcwd()


def volumes_path():
    # volumes路径
    return '/srv/drops/%s/volumes/' % config.Conf().getProjectName()


def container_path():
    # 容器路径
    return '/srv/drops/%s/' % config.Conf().getProjectName()


def servers_path():
    return container_path() + 'servers/'


def docker_path():
    return container_path() + 'docker-compose.yaml'


def release_path():
    # 发布路径
    return container_path() + 'release/'


def var_path():
    return container_path() + 'var/'


def docker_cmd_template(cmd):
    # 执行 docker-compose 命令的模板，看起来像是这样：
    #  cd <container_path> && docker-compose %s"
    return 'cd ' + container_path() + ' && docker-compose %s' % cmd


def ssh_template_key(key_path, port, username, host, b=''):
    # ssh 登录运行命令
    return 'ssh -p %d -i %s %s@%s "cd %s && %s"' % (port, key_path, username, host, container_path(), b)


def ssh_template_pwd(password, port, username, host, b=''):
    # ssh 登录运行命令
    return 'sshpass -p %s ssh -p %d %s@%s "cd %s && %s"' % (password, port, username, host, container_path(), b)


def rsync2remotely(host, src, target):
    # rsync 本地同步到远程路径
    if host.key:
        b = 'rsync -avzP --del -e "ssh -p {port} -i {key_path}" --exclude "drops.yaml" --exclude ".git" --exclude ".gitignore" {src} {username}@{host}:{target}'
        echo_b = b.format(
            src=src, target=target, key_path='<key_path>', port=host.port, username=host.username, host=host.host)
        b = b.format(
            src=src, target=target, key_path=host.key, port=host.port, username=host.username, host=host.host)
    else:
        detection_cmd('sshpass')
        b = 'sshpass -p {password} rsync -avzP --del -e "ssh -p {port}" --exclude "drops.yaml" --exclude ".git" --exclude ".gitignore" {src} {username}@{host}:{target}'
        echo_b = b.format(
            src=src, target=target, password='<password>', port=host.port, username=host.username, host=host.host)
        b = b.format(
            src=src, target=target, password=host.password, port=host.port, username=host.username, host=host.host)
    print(echo_b)
    os.system(b)


def rsync2local(host, src, target):
    # rsync 远程同步到本地
    if host.key:
        b = 'rsync -avzP --del -e "ssh -p {port} -i %s" {username}@{host}:{src} {target}'.format(
            src=src, target=target, port=host.port, username=host.username, host=host.host)
        print(b%('<key_path>'))
        b = b % host.key
    else:
        detection_cmd('sshpass')
        b = 'sshpass -p %s rsync -avzP --del -e "ssh -p {port}" {username}@{host}:{src} {target}'.format(
            src=src, target=target,  port=host.port, username=host.username, host=host.host)
        print(b%('<password>'))
        b = b % host.password
    os.system(b)


def rsync2local_link_dest(host, src, target, link_dest):
    # rsync 远程同步到本地，设置链接
    if host.key:
        b = 'rsync -avzP --del -e "ssh -p {port} -i %s" --link-dest={link_dest} {username}@{host}:{src} {target}'.format(
            src=src, target=target,  port=host.port, username=host.username, host=host.host, link_dest=link_dest)
        print(b)
        b = b % host.key
    else:
        detection_cmd('sshpass')
        b = 'sshpass -p {%s} rsync -avzP --del -e "ssh -p {port}" --link-dest={link_dest} {username}@{host}:{src} {target}'.format(
            src=src, target=target,  port=host.port, username=host.username, host=host.host, link_dest=link_dest)
        print(b)
        b = b % host.password
    os.system(b)


def Fatal(*e):
    print("Fatal!", *e)


def add_ssh_arg(p):
    p.add_argument('-h', type=str, help='host', required=False)
    p.add_argument('-p', '--port', dest='port', type=int,
                   help="ssh port", default=22)


def add_container_arg(p):
    p.add_argument('container', type=str,
                   help="docker container name", nargs=1)


def add_arg_container(p):
    p.add_argument('container',
                   help="要执行操作的容器。", type=str, nargs='*')


def add_arg_force(p):
    p.add_argument("-f", '--force',
                   help="强制执行，不再提示确认。", default=False, action='store_true')


def add_arg_host(p):
    p.add_argument('-a', '--hostAlias',
                   help="要执行操作的主机别名。", type=str, default='default')


def get_arg_host_from_conf(p):
    return config.Conf().get_host(p.hostAlias)


def parse_args(p):
    return p.parse_args()


def detection_cmd(*bl):
    # 确认命令是否存在
    for b in bl:
        for p in os.environ['PATH'].split(':'):
            if os.path.isdir(p) and b in os.listdir(p):
                break
        else:
            raise er.CmdNotExist(b)
    return True


def ssh_shell(host, b):
    detection_cmd('ssh')
    if host.key:
        os.system(ssh_template_key(
            host.key, host.port, host.username, host.host, b))
    else:
        detection_cmd('sshpass')
        os.system(ssh_template_pwd(
            host.password, host.port, host.username, host.host, b))


def rsync_release(host, force):
    print('------- sync release -------')
    detection_cmd('rsync')
    # TODO 0.2.1 添加向下兼容
    if not os.path.isdir('release'):
        os.mkdir('release')
    rsync2remotely(host, 'release', container_path())


def rsync_docker(host, force=False):
    # 同步项目到远程目录
    print('------- sync docker-compose.yaml -------')
    detection_cmd('rsync')
    if not force and not confirm_drops_project(host):
        if not user_confirm("主机 %s 远程目录 %s 可能不是 drops 项目，是否继续同步？" % (host.host, container_path())):
            raise er.UserCancel
    rsync2remotely(host, 'docker-compose.yaml', docker_path())


def rsync_servers(host, force):
    print('------- sync servers -------')
    detection_cmd('rsync')
    rsync2remotely(host, 'servers', container_path())


def rsync_var(host, force):
    if not confirm_empty_dir(host, var_path()):
        if not force and not user_confirm("主机 %s 远程目录 %s 不是空目录，是否继续同步？" % (host.host, var_path())):
            raise er.UserCancel
    rsync2remotely(host, 'var', container_path())


def rsync_volumes(host, force):
    if not confirm_empty_dir(host, volumes_path()):
        if not force and not user_confirm("主机 %s 远程目录 %s 不是空目录，是否继续同步？" % (host.host, volumes_path())):
            raise er.UserCancel
    rsync2remotely(host, 'volumes', container_path())


def sync(host, force=False, obj='ops'):
    # rsync 不会创建上一层文件夹，同步前创建
    c = ssh.Client(**host.to_conf())
    _, status = c.exec(' mkdir -p ' + container_path())
    if status != 0:
        raise er.CmdExecutionError('mkdir -p ' + container_path() + ', code=' + str(status))
    arg = (host, force)
    if obj == 'ops':
        rsync_docker(*arg)
        rsync_release(*arg)
        rsync_servers(*arg)
    elif obj == 'docker':
        rsync_docker(*arg)
    elif obj == 'release':
        rsync_release(*arg)
    elif obj == 'servers':
        rsync_servers(*arg)
    elif obj == 'var':
        rsync_var(*arg)
    elif obj == 'volumes':
        rsync_volumes(*arg)
    else:
        raise er.UnsupportedSyncObject(obj)


def rsync_backup(host, src, target, link_desc=''):
    detection_cmd('rsync')
    if link_desc:
        return rsync2local_link_dest(host, src, target, link_desc)
    return rsync2local(host, src, target)


def backup(host, obj, target, time_format='%Y-%m-%d_%H:%M:%S', link_desc='', keep=-1, cod=False, force=False):
    backup2dir = target
    link_dir = link_desc
    if cod:  # 创建项目名的文件夹
        backup2dir = os.path.join(backup2dir, config.Conf().getProjectName())
    # 如果设定了时间格式
    if time_format:
        # 备份目录是备份对象下的时间目录
        backup2dir = os.path.join(backup2dir, '{obj}')
        # 如果设置了 link_desc
    # 需要备份的目录列表
    srcLi = []
    if obj == 'all':
        srcLi = [
            [release_path(), backup2dir.format(obj='release')],
            [servers_path(), backup2dir.format(obj='servers')],
            [var_path(), backup2dir.format(obj='var')],
            [volumes_path(), backup2dir.format(obj='volumes')],
            [docker_path(), backup2dir.format(obj='docker-compose.yaml')],
        ]
    elif obj == 'ops':
        srcLi = [
            [release_path(), backup2dir.format(obj='release')],
            [servers_path(), backup2dir.format(obj='servers')],
            [docker_path(), backup2dir.format(obj='docker-compose.yaml')],
        ]
    elif obj == 'docker':
        srcLi = [
            [docker_path(), backup2dir.format(obj='docker-compose.yaml')],
        ]
    elif obj == 'release':
        srcLi = [
            [release_path(), backup2dir.format(obj='release')],
        ]
    elif obj == 'servers':
        srcLi = [
            [servers_path(), backup2dir.format(obj='servers')],
        ]
    elif obj == 'var':
        srcLi = [
            [var_path(), backup2dir.format(obj='var')],
        ]
    elif obj == 'volumes':
        srcLi = [
            [volumes_path(), backup2dir.format(obj='volumes')],
        ]
    else:
        raise er.UnsupportedBackupObject(obj)

    for s in srcLi:
        to_path = s[1]
        if not os.path.isdir(s[1]):
            os.mkdir(s[1])
        if not link_desc:
            # 组装 link_desc
            lsdir = []
            # 排除不符合 format 的目录
            for i in os.listdir(s[1]):
                try:
                    time.strptime(i, time_format)
                    lsdir.append(i)
                except ValueError:
                    pass
            if not lsdir:
                print("%s 路径下没有找到符合 %s 的文件夹，--link-dest 功能关闭。" %
                      (s[1], time_format))
            # 备份文件夹的时间命名
            tar_time = time.strftime(
                time_format, time.localtime(time.time()))
            if tar_time in lsdir:
                lsdir.remove(tar_time)  # 排除掉本次要备份的文件夹
            if lsdir:
                lsdir.sort(reverse=True)
                # link 目录是排序后最大的目录
                link_dir = lsdir[0]
                link_desc = os.path.join(work_path(), s[1], link_dir)
                if keep > 0:
                    # 本次新增的备份也算一个
                    rm_dir = lsdir[keep-1:]
                    for r in rm_dir:
                        rd = os.path.join(s[1], r)
                        print('delete:', rd)
                        # docker-compose.yaml 是文件，其他是目录
                        if os.path.isdir(rd):
                            shutil.rmtree(rd)
                        else:
                            os.remove(rd)
                    if keep == 1:
                        link_desc = ''
            to_path = os.path.join(s[1], tar_time)
        if not force and os.path.isdir(to_path) and os.listdir(to_path):
            if not user_confirm("本地目录 %s 已存在且非空，继续同步可能会导致文件丢失，是否继续同步？" % (to_path)):
                raise er.UserCancel
        rsync_backup(host, s[0], to_path, link_desc)


def docker_compose_cmd(cmd, host):
    # docker_compose_cmd 解析命令行参数，只需要 ssh 部分的参数。
    # 接收一个字符串模板，
    # 拼接进入工作目录的命令，传入 cmd 并返回。
    # 可以执行任意 docker-compose 命令。
    for i in ('&', '`', '"', "'", ';'):  # 防止执行其他什么东西
        if i in cmd:
            raise er.CmdCannotContain(i)
    return exec(docker_cmd_template(cmd), host)


def exec(cmd, host):
    # 对 host 执行任意远程命令
    print('run host: ', host.host)
    c = ssh.Client(**host.to_conf())
    stdout, status = c.exec(cmd)
    if status != 0:
        print('----------------- fail -----------------')
        print(stdout, status)
        raise er.CmdExecutionError(cmd + ', code=' + str(status))
    return 0


def user_confirm(*l):
    s = ' '.join(l)
    for _ in range(3):
        try:
            a = input(s + '[y/N]')
        except KeyboardInterrupt:
            return False
        if a == 'y':
            return True
        elif a in ['n', 'N']:
            return False
    return False


def confirm_drops_project(host):
    # 防止出现同步时误删除，同步前检查目录。
    c = ssh.Client(**host.to_conf())

    # 目录存在返回0，否则返回1
    while True:
        _, s = c.exec(
            '''if [ -d %s ]; then exit 0; else exit 1; fi''' % container_path(), False)
        if s != -1:
            break
    # 目录不存在时可以安全同步
    if s == 1:
        return True
    b = '''if [ -f %s ] && [ -d %s ]; then exit 0; else exit 1; fi''' % (
        container_path()+'/docker-compose.yaml', container_path()+'/servers')

    while True:
        _, s = c.exec(b, False)
        # 这条命令很容易返回 -1，貌似是 paramiko 的锅。重试。
        if s != -1:
            break
    # 目录存在，却没有 docker-compose 或 servers，可能不是 drops 项目
    if s == 0:
        return True
    return False


def confirm_empty_dir(host, path):
    # 返回 path 是否是空目录
    c = ssh.Client(**host.to_conf())
    # 目录存在返回0，否则返回1
    while True:
        _, s = c.exec(
            '''if [ -d %s ]; then exit 0; else exit 1; fi''' % path, False)
        if s != -1:
            break
    # 目录不存在时可以安全同步
    if s == 1:
        return True
    # 空目录返回 0， 否则返回 1
    while True:
        _, s = c.exec(
            '''if [ "$(ls -A %s)" ]; then exit 1; else exit 0 ; fi''' % path, False)
        if s != -1:
            break

    if s == 0:
        return True
    return False



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


from . import internal


def add_init_env_debian_cmd(s):
    p = s.add_parser(
        'initDebianEnv', help='初始化 debian 系远程环境。')
    internal.add_arg_host(p)
    p.set_defaults(func=init_env_debian_cmd)


def init_env_debian_cmd(p):
    host = internal.get_arg_host_from_conf(p)
    bin = 'apt-get update && apt-get install -y rsync docker-compose'
    internal.exec(bin, host)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import os
import shutil

import drops
from .config import Conf


def add_init_cmd(s):
    p = s.add_parser(
        'init', help='在当前目录初始化项目。')
    p.set_defaults(func=new_init)
    p.add_argument("projectName", metavar="projectName",
                   type=str, help="项目名。", nargs='?')


def new_init(p):
    objPath = os.path.join(drops.__path__[0], 'docker_ops')
    pwd = os.getcwd()
    if p.projectName == None:
        p.projectName = os.path.split(pwd)[-1]

    for i in os.listdir(objPath):
        s = os.path.join(objPath, i)
        t = os.path.join(pwd, i)
        if os.path.isdir(s):
            shutil.copytree(s, t)
        else:
            shutil.copyfile(s, t)

    c = Conf()
    c.setProjectName(p.projectName)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


from . import internal


def add_echo_paths_cmd(s):
    p = s.add_parser(
        'echoPaths', help='显示 drops 用到的各部署路径')
    p.set_defaults(func=deploy_echo_paths_cmd)


def deploy_echo_paths_cmd(p):
    print('容器路径，docker-compose.yaml 所在文件夹:', internal.container_path())
    print('发布路径，应用程序文件夹:', internal.release_path())
    print('volumes 路径，应用程序数据文件夹:', internal.volumes_path())



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import os
import shutil

import drops


def add_new_cmd(s):
    p = s.add_parser(
        'new', help='Create a drops project.')
    p.add_argument("dirName",  type=str, help="目录名，如果没有项目名会作为默认项目名。")
    p.add_argument("projectName",  type=str,
                   help="项目名。", default='', nargs='?')
    p.set_defaults(func=new_cmd)


def new_cmd(arg):
    from . import config
    from . import internal

    if arg.dirName in os.listdir(os.getcwd()):
        internal.Fatal("File exists: ", os.path.join(os.getcwd(), arg.dirName))
        return
    objPath = drops.__path__[0]
    shutil.copytree(os.path.join(objPath, 'docker_ops'),
                    os.path.join(os.getcwd(), arg.dirName))
    os.chdir(arg.dirName)
    c = config.Conf().open()
    if arg.projectName:
        c.setProjectName(arg.projectName)
    else:
        c.setProjectName(arg.dirName)
    c.save()



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from .clean_up import *
from .undeploy import *

from .deploy import *
from .init_remove_env_debian import *
from .internal import *
from .new import *
from .nginx_force_reload import *
from .nginx_reload import *
from .ps import *
from .redeploy import *
from .sync import *
from .init import *
from .host import *
from .stop import *
from .rm import *
from .kill import *
from .up import *
from .start import *
from .restart import *
from .project import *
from .echo_paths import *
from .pull import *
from .deploy_https_key import *
from .logs import *
from .exec import *
from .backup import *


def initCmd(s):
    add_clean_up_cmd(s)
    add_undeploy_cmd(s)

    add_new_cmd(s)
    add_ps_cmd(s)
    add_init_cmd(s)
    add_host_cmd(s)
    add_sync_cmd(s)
    add_deploy_cmd(s)
    add_redeploy_cmd(s)
    add_init_env_debian_cmd(s)
    add_nginx_reload_cmd(s)
    add_nginx_force_reload_cmd(s)
    add_stop_cmd(s)
    add_rm_cmd(s)
    add_kill_cmd(s)
    add_up_cmd(s)
    add_start_cmd(s)
    add_restart_cmd(s)
    add_project_cmd(s)
    add_echo_paths_cmd(s)
    add_pull_cmd(s)
    add_deploy_https_key_cmd(s)
    add_logs_cmd(s)
    add_exec_cmd(s)
    add_backup_cmd(s)



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import os

# 避免出现 "ImportError: attempted relative import with no known parent package"。
try:
    from . import globals
except:
    import globals

# self.args 是类接收到的所有参数。这样写只是为了少些两行代码，忽视可维护和语法检查。


class DropsErr(Exception):
    pass


class ThisIsNotDropsProject(DropsErr):
    def __str__(self):
        cwd = os.path.split(os.getcwd())[-1]
        return '"%s" is not a drops project, drops.yaml does not exist.' % cwd


class RemotePathIsNotDropsProject(DropsErr):
    def __str__(self):
        return 'Remote host "%s" is not a drops project, docker-compose.yaml does not exist.' % self.args[0]


class UserCancel(DropsErr):
    def __str__(self):
        return "用户取消。"


class ConfigFileAlreadyExists(DropsErr):
    def __str__(self):
        return '%s already exists.' % globals.confFileName


class PwdAndKeyCannotBeEmpty(DropsErr):
    def __str__(self):
        return 'Password and key cannot be empty at the same time.'


class ArgsError(DropsErr):
    def __str__(self):
        return 'Args error: %s.' % (' '.join(self.args))

class UnsupportedSyncObject(DropsErr):
    def __str__(self) -> str:
        return 'Unsupported sync object: %s.' % (' '.join(self.args))


class UnsupportedBackupObject(DropsErr):
    def __str__(self) -> str:
        return 'Unsupported backup object: %s.' % (' '.join(self.args))


class CmdNotExist(DropsErr):
    def __str__(self) -> str:
        return '[%s] command does not exist, please install it manually.' % (', '.join(self.args))


class HostExisted(DropsErr):
    def __str__(self):
        return 'Host existed: %s.' % self.args[0]


class ConfigurationFileFormatError(DropsErr):
    def __str__(self):
        return 'Configuration file format error.'

class HostDoesNotExist(DropsErr):
    def __str__(self):
        return 'Host %s does not exist' % self.args[0]


class CmdCannotContain(DropsErr):
    def __str__(self):
        return 'Command cannot contain "%s".' % self.args[0]


class CmdExecutionError(DropsErr):
    def __str__(self):
        return 'Command execution error: "%s".' % self.args[0]



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.

from . import config


def add_project_cmd(s):
    p = s.add_parser(
        'project', help='项目信息')
    p.set_defaults(func=project_cmd)

    subparsers = p.add_subparsers(metavar="")
    n = subparsers.add_parser('name')
    n.add_argument('name', help="传入参数设置项目名，没有参数输出项目名",
                   type=str, nargs='?', default=None)


def project_cmd(p):
    c = config.Conf().open()
    if 'name' in p:
        if p.name == None:
            print(c.getProjectName())
        else:
            c.setProjectName(p.name)
    else:
        for k, i in c.C['project'].items():
            print(k, ': ', i, sep='')



# Copyright © 2022 szerr < i@szerr.org > .

# This file is part of drops.

# drops is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.

# drops is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with drops. If not, see <https://www.gnu.org/licenses/>.


import yaml
import os

from . import globals
from . import er


class Host():
    def __init__(self, host, port, username, password=None, key=None, coding='utf-8'):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.key = key
        self.coding = coding

    def __dict__(self):
        return {'host': self.host, 'port': self.port,
                'username': self.username, 'password': self.password,
                'key': self.key, 'coding': self.coding}

    def to_conf(self):
        t = {'host': self.host, 'port': self.port,
             'username': self.username, 'coding': self.coding}
        if self.password:
            t['password'] = self.password
        elif self.key:
            t['key'] = self.key
        else:
            raise er.PwdAndKeyCannotBeEmpty
        return t


class Conf():
    # 封装配置文件
    def __init__(self):
        self.C = None

    def open(self):
        if self.C == None:
            if not os.path.isfile(globals.confFileName):
                raise er.ThisIsNotDropsProject
            with open(globals.confFileName) as fd:
                self.C = yaml.load(fd.read(), Loader=yaml.Loader)
            if type(self.C) is not dict:
                raise er.ConfigurationFileFormatError
            if self.C.get('hosts', None) == None:
                self.C['hosts'] = {}
            return self

    def setProjectName(self, n):
        self.open()
        # 设置项目名
        p = self.C.get('project', {})
        p['name'] = n
        self.C['project'] = p
        self.save()
        return self

    def getProjectName(self):
        self.open()
        # TODO 0.1.6 之前项目名用的文件夹名，做个兼容。
        if 'project' not in self.C:
            n = os.path.split(os.getcwd())[-1]
            self.setProjectName(n)
            self.save()
        return self.C['project']['name']

    def save(self):
        with open(globals.confFileName, 'w') as fd:
            fd.write(yaml.dump(self.C))
        return self

    def new(self):
        if os.path.isfile(globals.confFileName):
            raise er.ConfigFileAlreadyExists
        self.C = {'hosts': {}}
        self.save()
        return self

    def add_host(self, hostAlias, host, port=22, username='root', password=None, key=None, coding='utf-8'):
        # 增加 主机配置。coding 为服务器 shell 编码
        self.open()
        h = Host(host, port, username, password, key, coding)
        if hostAlias in self.C['hosts']:
            raise er.ArgsError('hostAlias:', hostAlias, '已存在')
        self.C['hosts'][hostAlias] = h.to_conf()
        self.save()
        return self

    def change_host(self, hostAlias, host, port=22, username='root', password=None, key=None, coding='utf-8'):
        # 修改 host
        self.open()
        h = Host(host, port, username, password, key, coding)
        self.C['hosts'][hostAlias] = h.to_conf()
        self.save()
        return self

    def remove_host(self, hostAlias):
        # 移除 host
        self.open()
        if hostAlias in self.C['hosts']:
            self.C['hosts'].pop(hostAlias)
        else:
            raise er.HostDoesNotExist(hostAlias)
        self.save()
        return self

    def print_minlength(self, s, length):
        # 最小打印长度
        ind = 0
        while True:
            l = ind + length - len(s)
            if l < 0:
                ind += length
            else:
                ind += l
                break
        return s + ind * ' '

    def print_host(self, name, hostinfo, head=''):
        print(head + name + ':')
        for k, i in hostinfo.items():
            p = self.print_minlength(k, 12)
            if k == 'password':
                print(head + '  ' + p, ' *')
            else:
                print(head + '  ' + p, i)

    def print_hosts(self, hosts, head=''):
        print(head + 'hosts:')
        for h, item in hosts.items():
            self.print_host(h, item, head+'  ')

    def ls(self, host=None):
        # 输出 host
        self.open()
        if host is None:
            self.print_hosts(self.C['hosts'])
            return self

        self.print_host(host, self.C['hosts'][host])
        return self

    def get_host(self, host_type):
        # 获取单个 host
        self.open()
        if host_type in self.C['hosts']:
            return Host(**self.C['hosts'][host_type])
        raise er.HostDoesNotExist(host_type)
